# Title:	MT-AutoBan Library
# Summary:	Update a .htaccess file to ban addresses of junk comments/trackbacks
# Author:	Alan M. Carroll
# Version:	1.2.2
# Date:		16 Aug 2006
#
# Copyright 2005-2006 Alan M. Carroll

use strict;
# ----------------------------------------------------------------------------
package MT::Plugin::AutoBan;
our $self;
our $UPDATE_REQUIRED;

# The number of junk objects to load at one time
use constant DEFAULT_BLOCK_SIZE => 2000;

# Scan junk objects for IP addresses.
# Args: block_size -- number of DB records to load at a time
sub loadAddrs {
    require MT::Comment;
    require MT::TBPing;
    
    my $block_size = shift;
    $block_size = DEFAULT_BLOCK_SIZE if (!$block_size or $block_size < 1);
    
    my %addrs; # the map of address -> count # return value
    my %db_args = ( limit => $block_size ); # arguments to load from DB
    my %db_terms; # terms for DB load
    
    ## Extra vars for time based retrieval
    # The base time for time span calculations. This is left undef if timespan
    # is disabled, otherwise it is the epoch time adjusted by the timespan.
    my $tick;
    my $has_multi_tz = 0; # set if there is more than one TZ among the weblogs
    my @blog_tz; # sets IDs of weblogs with the same TZ
    
    my @objects; # current working set of objects
    
    # Generate the list of the types of objects we look through for junk.
    my @junk_types;
    push(@junk_types, 'MT::Comment') if $self->get_config_value('check_comments');
    push(@junk_types, 'MT::TBPing') if $self->get_config_value('check_tbs');
    
    # If timespan is active, generate the data we need for it.
    if (my $timespan = $self->get_config_value('check_timespan') && $self->get_config_value('timespan')) {
        # don't need this here, but will later in a loop so do it outside the loop.
        require MT::Util;
        
        # need to iterate over the weblogs and classify them by TZ
        require MT::Blog;
        my %tz; # map of weblog timezone offset to weblog IDs
        foreach my $blog (MT::Blog->load()) {
            my $dt = $blog->server_offset;
            if (exists $tz{$dt}) {
                push(@{$tz{$dt}}, $blog);
            } else {
                $tz{$dt} = [ $blog ];
            }
        }
        # Now assemble the distinct lists of weblog IDs
        @blog_tz = values %tz;
        # Special case when all weblogs have the same TZ
        $has_multi_tz = $#blog_tz > 0;
        # Because timespan is enabled we need to sort every time
        $db_args{sort} = 'created_on';
        
        # Set base time for later use.
        $tick = time() - 60 * $timespan;
    }

    # gather addresses from junk trackbacks and comments
    do {
        # If checking time, adjust DB query data
        if (my $set = pop @blog_tz) {
            # We can just use the first weblog on the list because they
            # all have the same TZ (that's why we classified them)
            $db_args{start_val} = MT::Util::epoch2ts($set->[0], $tick);
            # If multiple TZs, then we need to specify the weblogs
            $db_terms{blog_id} = [ map $_->id, @$set ] if $has_multi_tz;
        }
        
        for my $type (@junk_types) {
            no strict 'refs'; # used every loop and outside, so turn it off here
            
            my $JUNK_METH = $type . '::JUNK'; # constant with the value for JUNK
            $db_terms{junk_status} = &$JUNK_METH;
        
            my $offset = 0;

            # Do the loads from the DB
            while (1) {
                $db_args{offset} = $offset;
                my @objects = $type->load( \%db_terms , \%db_args );
                last if not @objects; # no records found, done.

                # Extract the IP addresses.
                $addrs{$_->ip} += 1 foreach @objects;

                if (scalar @objects < $block_size) {
                    last; # partial block -> no more records
                } else {
                    $offset += $block_size; # bump offset to get next block
                }
            }
        }
    } while (scalar @blog_tz);

    return \%addrs;
}

# Update the .htaccess file for the current set of banned addresses
sub update_access_file {
    require File::Spec->catdir(MT->instance->mt_dir, $self->envelope, 'IPAccess.pm');
    
    my $threshold = $self->get_config_value('threshold');
    my $block_size = $self->get_config_value('block_size') || DEFAULT_BLOCK_SIZE;
    my $junk_count = 0; # total # of junk objects
    my $addr_count = 0; # number of distinct addresses found

    my $access_file = new Transfinitum::Apache::IPAccess( $self->get_config_value('path')
                                                        , no_load => 1
                                                        , lock => 1
                                                        );
    # If new fails, it means we didn't get the lock. That means some other
    # instance is updating the file, so it's best if we just quit
    # immediately.
    return if not $access_file;
    
    # Get addresses from junk objects
    my $addrs = loadAddrs($block_size);
    
    # Find the addresses to be banned and gather statistics as long as we're
    # in the neighborhood.
    while ( my ($addr, $count) = each %$addrs) {
        $junk_count += $count;  # number of junk objects
        $addr_count += 1;       # number of unique addresses
        $access_file->ban($addr) if $count >= $threshold;
    }
    
    # Do the actual file update now.
    $access_file->{is_modified} = 1;
    my $ban_count = $access_file->count();
    my $status = "Updated " . localtime() . "<br />$junk_count Junk objects ("
                              . ($self->get_config_value('check_comments') ? 'C' : '')
                              . ($self->get_config_value('check_tbs') ? 'T' : '')
                              .  ") with $addr_count unique addresses. $ban_count banned at threshold $threshold."
                              ;
    my @update_args = (tag => $status);
    push(@update_args, url => $self->get_config_value('deny_url')) if $self->get_config_value('deny_type');
    push(@update_args, rewrite => 1) if $self->get_config_value('deny_type') == 2;
    push(@update_args, per_line => $self->get_config_value('addr_per_line')) if $self->get_config_value('addr_per_line');
    $access_file->update(@update_args);
    
    # Store the statistics for retrieval
    $self->set_config_value('status', $status);
    stats_set( last_updated => time()
             , object_count => $junk_count
             , banned => $ban_count
             , unique => $addr_count
             );
}

# Generate the HTML for the Plugin Settings page.
sub generate_config_html {
    my ($plugin, $param) = @_;
    no strict 'refs';
    # Is it worth making these package vars just to set the programmatically
    our ($path, $threshold, $timespan, $check_comments, $check_tbs, $check_timespan, $deny_url, $deny_type, $block_size, $addr_per_line);
    
    # Extract the parameters for re-use in the HTML generation.
    # Flags
    foreach (qw(check_comments check_timespan check_tbs)) {
        ${$_} = $param->{$_} ? 'checked="checked"' : '';
    }
    # Values
    foreach (qw(path threshold timespan deny_url deny_type block_size addr_per_line)) {
        ${$_} = $param->{$_} || 0;
    }
    
    my $deny_type_html = '<select name="deny_type" id="deny_type">';
    for my $k (0..2) {
        $deny_type_html .= '<option value="' . $k . '"';
        $deny_type_html .= ' selected="1"' if $k == $deny_type;
        $deny_type_html .= '>' . qw(None ErrorDocument mod_rewrite)[$k] . '</option>';
    }
    $deny_type_html .= '</select>';
    my $header_style = 'style="background:#EEE;font-weight:bold;border-top:2px solid #888;padding-left:0ex;padding-right:1ex;padding-top:2px;padding-bottom:2px;margin:0px;margin-bottom:1ex;"';
    my $separator = '<div style="margin:0px;clear:left;">&nbsp;</div>';
    # Need this to have nice displays on different browsers
    my $inter_field = '<div style="margin-bottom:2px;">&nbsp;</div>';
    
    my $status = 'Access file updated ' . stats_last_updated() . '.<br />'
               . 'Banned ' . stats_banned() . ' addresses out of ' . stats_unique() . ' unique addresses in ' . stats_object_count() . ' junk objects.';
    
    # Generate the HTML
    my $tmpl = <<TMPL;
    <div class="setting">
        <div $header_style>Basic settings.</div>
        <label for="threshold">Ban address if </label>
        <input type="text" name="threshold" id="threshold" size="2" value="$threshold"/> or more junk objects from that address.
        $inter_field
        <label for="check_comments">Check </label>
        <input type="checkbox" name="check_comments" id="check_comments" value="true" $check_comments /> Junk Comments
        <br />
        <label for="check_tbs">Check </label>
        <input type="checkbox" name="check_tbs" id="check_tbs" value="true" $check_tbs /> Junk Trackbacks
        <br />
        <label for="timespanflag">Check </label>
        <input type="checkbox" name="check_timespan" id="check_timespan" value="true" $check_timespan />
            Only junk objects less than
            <input type="text" name="timespan" id="timespan" size="5" value="$timespan" /> minutes old.
        $separator
        <div $header_style>Expert settings. These are useful (sometimes) for people familiar with access files.</div>
        <label for="path">Access File Path: </label>
        <input type="text" name="path" id="path" size="80%" value="$path" />
        <p style="margin-left:2ex;"><em>Relative paths are relative to the MT install directory</em></p>
        $inter_field
        <label for="deny_type">Redirection: </label> $deny_type_html
        <p style="margin-left:2ex;"><em>Banned addresses can be redirected to a specific URL either by an ErrorDocument directive or using mod_rewrite</em></p>
        $inter_field
        <label for="deny_url">Ban page URL: </label>
        <input type="text" name="deny_url" id="deny_url" size="80%" value="$deny_url" />
        $separator
        <div $header_style>Performance tuning parameters</div>
        <label for="block_size">Read</label>
        <input type="text" name="block_size" id="block_size" size="5" value="$block_size"/> objects at a time from the database.
        $inter_field
        <label for="addr_per_line">Write </label>
        <input type="text" name="addr_per_line" id="addr_per_line" size="5" value="$addr_per_line"/> addresses per line in the access file.
        $separator
        <div $header_style>Actions to take upon save.</div>
        <label for="force_refresh"></label>
        <input type="checkbox" name="force_refresh" id="force_refresh" /> Update access file
        $separator
        <div $header_style>Status</div>
        <p>$status</p>
    </div>
TMPL
    $tmpl;
}

# Override configuration save so we do fixups and special checks
sub _save_config {
    my $self = shift;
    my $param = $_[0]; # must preserve this for chaining
    
    # Check for force update and clear it because we don't save it.
    $UPDATE_REQUIRED = 1 if delete $param->{force_refresh};
    
    # Fix up file path
    my $path = $param->{path};
    if (not $path) {
        $param->{path} = $self->{settings}->defaults()->{path};
    } elsif ($path =~ m!/$!) {
        $param->{path} .= '.htaccess';
    } elsif ($path =~ m!^\.\.?$!) {
        $param->{path} .= '/.htaccess';
    }
    
    # Fix up threshold
    my $threshold = $param->{threshold};
    if (not $threshold or $threshold !~ m!^[[:space:]]*[[:digit:]]+[[:space:]]*$!) {
        $param->{threshold} = $self->{settings}->defaults()->{threshold};
    }
    
    # Now chain up to the real method to do the actual saving.
    $self->SUPER::save_config(@_);
}
# ----------------------------------------------------------------------------
# Store various statistics for display in the settings or via tags.

our $STATS_DATA;
# last_updated -> Epoch time of last update, UTC
# object_count -> number of junk objects checked
# unique -> number of unique addresses
# banned -> number of banned addresses
sub obtain_stats {
    require MT::PluginData;
    
    return $STATS_DATA if $STATS_DATA;
    return $STATS_DATA if $STATS_DATA = MT::PluginData->load( { plugin => 'MTAutoBan', key => 'stats' });
    # Not there, create a new one
    $STATS_DATA = MT::PluginData->new;
    $STATS_DATA->plugin('MTAutoBan');
    $STATS_DATA->key('stats');
    $STATS_DATA->data({ last_updated => 0 , object_count => 0 , unique => 0 , banned => 0 });
    $STATS_DATA;
}

sub stats_last_updated {
    my ($ctx, $args, $cond) = @_;
    my $last_updated = obtain_stats()->data()->{last_updated};
    return 'Never' if not $last_updated;
    require MT::Util;
    my $blog = $ctx && $ctx->stash('blog');
    # because format_ts uses the weblog TZ, it does its base computations
    # using UTC. So if we don't have a weblog in context, we need to
    # adjust using the system TZ.
    if (not $blog) {
        use Time::Local;
        $last_updated = timegm(localtime($last_updated));
    }
    return MT::Util::format_ts($args && $args->{format}, MT::Util::epoch2ts($blog, $last_updated));
}

sub stats_object_count { return obtain_stats()->data()->{object_count}; }
sub stats_unique { return obtain_stats()->data()->{unique}; }
sub stats_banned { return obtain_stats()->data()->{banned}; }
sub stats_set {
    my $stats = obtain_stats();
    my $fields;
    if (ref $_[0] eq 'HASH') {
        $fields = shift;
    } else {
        $fields = { @_ };
    }
    $stats->data($fields);
    $stats->save();
}
# ----------------------------------------------------------------------------
1;

__END__

## Debugging stuff I don't want to rewrite in the future.

sub _dlog {
    my $file;
    open $file, '>>', 'mt-log.txt';
    print $file map { if (ref $_ eq 'ARRAY') { ( '(', join(', ', @$_), ')' ) } else { $_ } } @_;
    close $file;
}

#                _dlog('Loaded ', scalar(@objects), ' ', $type, "\n");
#                _dlog("Terms\n", map { ( '  ', $_, '=', $db_terms{$_}, "\n" ) } keys(%db_terms));
#                _dlog("Args\n", map { ( '  ', $_, '=', $db_args{$_}, "\n" ) } keys(%db_args));
