# Title:	MT-AutoBan
# Summary:	Update a .htaccess file to ban addresses of junk comments/trackbacks
# Author:	Alan M. Carroll
# Version:	1.2.3
# Date:		07 Sep 2006
#
# Copyright 2005-2006 Alan M. Carroll

use strict;
# ----------------------------------------------------------------------------
package MT::Plugin::AutoBan;
use base qw(MT::Plugin);
use MT;
use MT::Template::Context;

our $VERSION = '1.2.4';

# This gets set to the path to our main library file during install
our $LIB;

# This gets set if an object changed junk status.
our $UPDATE_REQUIRED = 0;

# This is a singleton, so keep track of it (set by install)
our $self;

# Call back on post save of an object. We can't really detect a transition
# (and the HandleJunk callback is not invoked when a comment is created)
# so we just update the bans whenever a junked comment is saved.
# (When a comment is unjunked, we get the HandleNotJunk callback)
sub on_save {
    my ($context, $obj) = @_;
    $UPDATE_REQUIRED = 1 if $obj->is_junk();
}

# Called when an object is marked not junk.
sub force_update {
    $UPDATE_REQUIRED = 1;
}

sub on_app_shutdown {
    require "$LIB";
    &update_access_file if $UPDATE_REQUIRED;
    $UPDATE_REQUIRED = 0; # anticipating mod_perl
}

# Override configuration save so we do fixups and special checks
# This must go in the plugin file or it won't be called reliably
sub save_config { require "$LIB"; &_save_config; }

# Register our plugin
sub install {
    my $plugin = MT::Plugin::AutoBan->new(
        { name => 'AutoBan'
        , description => 'Automatically ban IP addresses of junk comments and trackbacks'
        , author_link => 'http://blog.thought-mesh.net/solidwallofcode'
        , author_name => 'Alan M. Carroll'
        , doc_link   => 'http://blog.thought-mesh.net/solidwallofcode/mt_projects/autoban.php'
        , version => $VERSION
        , settings => new MT::PluginSettings(
            [ [ threshold => { Default => 2 } ]
            , [ check_comments => { Default => 1 } ]
            , [ check_tbs => { Default => 1 } ]
            , [ path => { Default => '.htaccess' } ]
            , [ timespan => { Default => 1440 } ]
            , [ check_timespan => { Default => 0 } ]
            , [ block_size => { Default => 2000 } ]
            , [ addr_per_line => { Default => 50 } ]
            , [ deny_type => { Default => 0 } ]
            , [ deny_url => { Default => '/sand-trap.php' } ]
            ]
        )
        , system_config_template => sub { require "$LIB"; &generate_config_html; }
        }
    );
    our $self = $plugin;
    MT->add_plugin($plugin);

    require MT::Comment;
    require MT::TBPing;

    MT::Comment->add_callback('post_save', 5, $plugin, \&on_save);
    MT::TBPing->add_callback('post_save', 5, $plugin, \&on_save);
    MT::Comment->add_callback('pre_remove', 5, $plugin, \&on_save);
    MT::TBPing->add_callback('pre_remove', 5, $plugin, \&on_save);
    MT->add_callback('TakeDown', 5, $plugin, \&on_app_shutdown);
    MT->add_callback('HandleNotJunk', 5, $plugin, \&force_update);
    MT->add_callback('HandleJunk', 5, $plugin, \&force_update);

    MT::Template::Context->add_tag( AutoBanObjectCount => sub { require "$LIB"; &stats_object_count(); } );
    MT::Template::Context->add_tag( AutoBanUniqueCount => sub { require "$LIB"; &stats_unique(); } );
    MT::Template::Context->add_tag( AutoBanBannedCount => sub { require "$LIB"; &stats_banned(); } );
    MT::Template::Context->add_tag( AutoBanLastUpdated => sub { require "$LIB"; &stats_last_updated(@_); } );
    MT::Template::Context->add_tag( AutoBanThreshold => sub { $self->get_config_value('threshold'); } );

    $LIB = File::Spec->catdir(MT->instance->mt_dir, $plugin->envelope, 'autoban-lib.pm');
}
# ----------------------------------------------------------------------------
package main;
MT::Plugin::AutoBan::install();
# ----------------------------------------------------------------------------
1;
